package ap.proof.goal;


object EagerTaskManager {

  /**
   * It is unknown whether the facts of the current goal are normalised
   */
  private object NonNormalisedFacts extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case FactsNormalisationTask => NormalisedFacts
      case _ =>                      NonNormalisedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = npt match {
      case Some(_ : BetaFormulaTask | _ : ExQuantifierTask | _ : DivisibilityTask) | None =>
        Some(FactsNormalisationTask)
      case _ =>
        None
    }
  }

  /**
   * It is known that <code>FactsNormalisationTask</code> has been applied, the
   * facts of the current goal are normalised
   */
  private object NormalisedFacts extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case _ : AddFactsTask =>   NonNormalisedFacts
      case UpdateTasksTask =>    NormalisedFactsAndTasks 
      case _ =>                  NormalisedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = npt match {
      case Some(_ : BetaFormulaTask | _ : ExQuantifierTask | _ : DivisibilityTask) | None =>
        Some(UpdateTasksTask)
      case _ =>
        None
    }
  }

  /**
   * It is known that <code>FactsNormalisationTask</code> has been applied, the
   * facts of the current goal are normalised, and also
   * <code>UpdateTasksTask</code> has been applied so that all tasks are
   * normalised
   */
  private object NormalisedFactsAndTasks extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case _ : AddFactsTask =>   NonNormalisedFacts
      case EliminateFactsTask => ReducedFacts
      case _ =>                  NormalisedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = npt match {
      case Some(_ : BetaFormulaTask | _ : ExQuantifierTask) | None =>
        Some(EliminateFactsTask)
      case _ =>
        None
    }
  }

  /**
   * It is known that <code>FactsNormalisationTask</code> and
   * <code>EliminateFactsTask<code> have been applied, but it could have
   * happened that afterwards constants have disappeared from the tasks in the
   * queue, so that further eliminations might be possible
   */
  private object ProbablyReducedFacts extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case FactsNormalisationTask | EliminateFactsTask => ReducedFacts
      case _ : AddFactsTask =>                            NonNormalisedFacts
      // all other tasks could result in the disappearance of constants in
      // the task queue, which could make it necessary to apply
      // <code>EliminateFactsTask</code> again
      case _ =>                                           ProbablyReducedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = npt match {
      case Some(_ : ExQuantifierTask | _ : DivisibilityTask) | None =>
        Some(EliminateFactsTask)
      case _ =>
        None
    }
  }

  /**
   * It is known that both <code>FactsNormalisationTask</code> and
   * <code>EliminateFactsTask<code> have been applied, the
   * facts of the current goal are normalised and unnecessary facts have been
   * removed (with respect to the current tasks)
   */
  private object ReducedFacts extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case FactsNormalisationTask | EliminateFactsTask => ReducedFacts
      case _ : AddFactsTask =>                            NonNormalisedFacts
      case OmegaTask =>                                   PreFinal
      // all other tasks could result in the disappearance of constants in
      // the task queue, which could make it necessary to apply
      // <code>EliminateFactsTask</code> again
      case _ =>                                           ProbablyReducedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = npt match {
      case None => Some(OmegaTask)
      case _ => None
    }
  }

 private object PreFinal extends EagerTaskManager {
   def afterTask(task : Task) = task match {
     case FactsNormalisationTask | EliminateFactsTask => ReducedFacts
     case SplitEqTask =>                                 Final
     case OmegaTask =>                                   PreFinal
     case _ : AddFactsTask =>                            NonNormalisedFacts
     // all other tasks could result in the disappearance of constants in
     // the task queue, which could make it necessary to apply
     // <code>EliminateFactsTask</code> again
     case _ =>                                           ProbablyReducedFacts
   }
   def recommend(npt : Option[PrioritisedTask]) = npt match {
     case None => Some(SplitEqTask)
     case _ => None
   }
 }

  /**
   * The final state in where there is nothing left to do. This state can be
   * reached by applying the task <code>SplitEqTask</code>. In case that actually an
   * equation was split, this will lead to <code>AddFactsTask</code>, and the
   * state will be left again immediately
   */
  private object Final extends EagerTaskManager {
    def afterTask(task : Task) = task match {
      case FactsNormalisationTask | EliminateFactsTask => ReducedFacts
      case SplitEqTask =>                                 Final
      case _ : AddFactsTask =>                            NonNormalisedFacts
      // all other tasks could result in the disappearance of constants in
      // the task queue, which could make it necessary to apply
      // <code>EliminateFactsTask</code> again
      case _ =>                                           ProbablyReducedFacts
    }
    def recommend(npt : Option[PrioritisedTask]) = None    
  }
   
  /**
   * In the beginning, there are no facts, which are thus reduced
   */
  def INITIAL : EagerTaskManager = Final
}


/**
 * A class for tracking the application of tasks and recommending the
 * intermediate application of <code>EagerTask</code>s. This class is
 * implemented as a finite automaton to give recommendations based on the
 * history of task applications
 */
abstract class EagerTaskManager {

  /**
   * Tell the manager that a certain task was applied
   */
  def afterTask(task : Task) : EagerTaskManager

  /**
   * Obtain a recommendation from the manager, given the next
   * <code>PrioritisedTask</code> in the queue. If the queue is empty,
   * <code>None</code> should be given as argument
   */
  def recommend(nextPrioritisedTask : Option[PrioritisedTask]) : Option[EagerTask]
  
}
