package ap.proof.goal;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.{Formula, ConstantTerm, VariableTerm,
                  TermOrder, TerFor, Term}
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions,
                               ReduceWithConjunction, Quantifier}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.preds.{PredConj, Atom}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.substitutions.{Substitution, ConstantSubst, IdentitySubst}
import ap.util.{Debug, FilterIt, Logic, Seqs}
import ap.parameters.{GoalSettings, Param}

object Goal {
  
  private val AC = Debug.AC_GOAL

  def apply(facts : Conjunction, clauses : NegatedConjunctions,
            tasks : TaskManager, age : Int,
            eliminatedConstants : Set[ConstantTerm], vocabulary : Vocabulary,
            definedSyms : Substitution, settings : GoalSettings) : Goal =
    if (facts.isFalse || clauses.isFalse)
      TRUE(vocabulary)
    else
      new Goal(facts, clauses, tasks, age, eliminatedConstants,
               vocabulary, definedSyms, settings)

  def apply(f : Formula, order : TermOrder, settings : GoalSettings) : Goal =
    apply(f, Set.empty, order, settings)
  
  def apply(f : Formula,
            eliminatedConstants : Set[ConstantTerm],
            order : TermOrder,
            settings : GoalSettings) : Goal = {
    val tasks = formulaTasks(Conjunction.conj(f, order), 0, settings)
    apply(Conjunction.TRUE,
          NegatedConjunctions.TRUE,
          TaskManager.EMPTY ++ tasks,
          0,
          eliminatedConstants,
          Vocabulary(order),
          new IdentitySubst (order),
          settings)
  }
  
  def TRUE(vocabulary : Vocabulary) : Goal =
    new Goal (Conjunction.FALSE,
              NegatedConjunctions.TRUE,
              TaskManager.EMPTY, 0,
              Set.empty, vocabulary,
              new IdentitySubst (vocabulary.order),
              GoalSettings.DEFAULT)
  
  val TRUE : Goal = TRUE(Vocabulary.EMPTY)
  
  /**
   * Create the tasks to store and handle an arbitrary given formula. This
   * method is part of the goal because different tasks might be created
   * depending on the settings
   */
  private def formulaTasks(formula : Conjunction, age : Int,
                           settings : GoalSettings) : Seq[FormulaTask] =
    if (formula.isFalse) {
      List()
    } else if (formula.isTrue || formula.isLiteral) {
      List(new AddFactsTask(formula, age))
    } else if (formula.isNegatedConjunction) {
      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(Goal.AC, disj.quans.isEmpty)
      //////////////////////////////////////////////////////////////////////////

      (for (f <- disj.negatedConjs; g <- formulaTasks(f, age, settings)) yield g) ++
      (if (disj.arithConj.isTrue && disj.predConj.isTrue) {
         List()
       } else {
         val facts = Conjunction(List(), disj.arithConj, disj.predConj,
                                 NegatedConjunctions.TRUE, formula.order).negate
         List(new AddFactsTask(facts, age))})
    } else if (formula.quans.isEmpty) {
      List(new BetaFormulaTask(formula, age, Param.SYMBOL_WEIGHTS(settings)))
    } else formula.quans.last match {
      case Quantifier.ALL => List(new AllQuantifierTask(formula, age))
      case Quantifier.EX =>
	if (formula.isDivisibility)
	  List(new DivisibilityTask(formula, age))
        else if (Param.POS_UNIT_RESOLUTION(settings) &&
                 (NegLitClauseTask isCoveredFormula formula))
	  List(new NegLitClauseTask(formula, age))
	else
	  List(new ExQuantifierTask(formula, age))
    }

}

class Goal private (val facts : Conjunction,
                    val clauses : NegatedConjunctions,
                    val tasks : TaskManager,
                    val age : Int,
                    val eliminatedConstants : Set[ConstantTerm],
                    val vocabulary : Vocabulary,
                    val definedSyms : Substitution,
                    val settings : GoalSettings) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Goal.AC,
                   // facts are really just literals (equations, inequalities,
                   // predicate literals)
                   facts.quans.isEmpty &&
                   facts.negatedConjs.isEmpty &&
                   // clauses are handled similar to facts; there are two
                   // kinds: quantified formulae that are instantiated
                   // through unit resolution with facts, and ground formulae
                   // that do not contain predicates or eliminated constants
                   // (splitting is unnecessary for such formulae)
                   clauses.forall((c) =>
                                  (NegLitClauseTask isCoveredFormula c) ||
                                  constantFreedom.isShielded(c, bindingContext) ||
                                  c.predicates.isEmpty &&
                                  Seqs.disjoint(eliminatedConstants, c.constants) &&
                                  Conjunction.collectQuantifiers(c).isEmpty) &&
                   !clauses.containsLiteral &&
                   !clauses.containsNegatedConjunction &&
                   // there should not be any free variables (use constants
                   // instead)
                   facts.variables.isEmpty && clauses.variables.isEmpty &&
                   // if a contradiction is detected, we can forget about
                   // everything
                   (!facts.isFalse || tasks.isEmpty && clauses.isEmpty) &&
                   // we must choose an ordering such that the constants to be
                   // eliminated are maximal
                   (facts isSortedBy order) &&
                   (clauses isSortedBy order) &&
                   (definedSyms isSortedBy order) &&
                   (order constantsAreMaximal eliminatedConstants))
  //////////////////////////////////////////////////////////////////////////////

  private def elimConstants(tfs : Seq[LinearCombination]) : Seq[LinearCombination] = {
    val res = new ArrayBuffer[LinearCombination]
    for (tf <- tfs) {
      if (!tf.isEmpty && !(this eliminates tf.leadingTerm)) {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(Goal.AC, Seqs.disjoint(tf.constants, eliminatedConstants))
        ////////////////////////////////////////////////////////////////////////
        res += tf
      }
    }
    res
  }
  
  private def elimConstants(tfs : Iterator[LinearCombination])
                                                : Iterator[LinearCombination] =
    FilterIt(tfs, (tf:LinearCombination) => {
      val res = tf.isEmpty || !(this eliminates tf.leadingTerm)
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(Goal.AC,
                       res == Seqs.disjoint(tf.constants, eliminatedConstants))
      //////////////////////////////////////////////////////////////////////////
      res
    })
    
  private def arithClauses : NegatedConjunctions =
    NegatedConjunctions(for (c <- clauses if (c.predicates.isEmpty)) yield c, order)
    
  lazy val closingConstraint : Conjunction = {
    val ac = facts.arithConj
    
    val selectedAC =
      if (Seqs.disjoint(eliminatedConstants, facts.predConj.constants)) {
        // only take literals that do not contain eliminated constants
        ArithConj(ac.positiveEqs.updateEqsSubset(elimConstants(ac.positiveEqs), order),
                  ac.negativeEqs.updateEqsSubset(elimConstants(ac.negativeEqs), order),
                  InEqConj(elimConstants(ac.inEqs.elements ++
                                         ac.inEqs.geqZeroInfs.elements),
                           order),
                  order)
    } else {
      // otherwise, we have to consider all arithmetic literals
      ac
    }
    
    constantFreedom.unshieldedPart(
      definedSyms(Conjunction(List(), selectedAC, PredConj.TRUE, arithClauses,
                              order)).negate,
      bindingContext)
  }
  
  lazy val closingConstantFreedom : ConstantFreedom =
    constantFreedom.findNonFreeness(closingConstraint, bindingContext)

  lazy val fixedConstantFreedom : Boolean =
    constantFreedom == closingConstantFreedom

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return whether <code>t</code> is a constant that is eliminated in this goal
   */
  def eliminates(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  lazy val reduceWithFacts : ReduceWithConjunction =
    ReduceWithConjunction(facts, order)
   
  //////////////////////////////////////////////////////////////////////////////

  val stepPossible : Boolean = !tasks.isEmpty

  val stepMeaningful : Boolean = stepPossible
  
  def step(ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Goal.AC, this.stepPossible)
    ////////////////////////////////////////////////////////////////////////////
    
    // apply the first task to the goal and then stop immediately
    val task = tasks.max
//    ap.util.Timer.measure(task.getClass.getName) {
      task(this, ptf)
//    }
  }

  /**
   * Create the tasks to store and handle an arbitrary given formula. This
   * method is part of the goal because different tasks might be created
   * depending on the settings
   */
  def formulaTasks(formula : Conjunction) : Seq[FormulaTask] =
    Goal.formulaTasks(formula, age, settings)
      
  def updateConstantFreedom(cf : ConstantFreedom) : Goal =
    if (constantFreedom == cf) {
      this
    } else {
      // we might have to re-activate clauses that are no longer shielded

      def illegalClause(c : Conjunction) =
        c.predicates.isEmpty &&
        !Seqs.disjoint(c.constants, eliminatedConstants) &&
        !cf.isShielded(c, bindingContext)
      val (otherStuff, realClauses) = clauses partition (illegalClause _)
      val newClauses = clauses.update(realClauses, order)

      val newTasks = for (c <- otherStuff; t <- formulaTasks(c)) yield t

      // rematch literals and clauses that contain constants whose freedom has
      // changed
      
      val changedConsts = constantFreedom changedConstants cf
      
      val (unchangedPosLits, changedPosLits) =
        Seqs.split(facts.predConj.positiveLits,
                   (l:Atom) => Seqs.disjoint(changedConsts, l.constants))
      val (unchangedNegLits, changedNegLits) =
        Seqs.split(facts.predConj.negativeLits,
                   (l:Atom) => Seqs.disjoint(changedConsts, l.constants))
      val unchangedLits =
        facts.predConj.updateLitsSubset(unchangedPosLits, unchangedNegLits, order)
      val changedLits =
        facts.predConj.updateLitsSubset(changedPosLits, changedNegLits, order)
        
      val (unchangedClauses, changedClauses) =
        (newClauses filter (NegLitClauseTask.isCoveredFormula _))
                   .partition ((c) => Seqs.disjoint(changedConsts, c.constants))
      
      val updatedVocabulary = vocabulary updateConstantFreedom cf
      val newTasks2 = NegLitClauseTask.rematchClauses(facts, updatedVocabulary,
                                                      unchangedLits, changedLits,
                                                      unchangedClauses, changedClauses,
                                                      this)
      
      Goal(facts, newClauses, tasks ++ newTasks ++ newTasks2, age,
           eliminatedConstants, updatedVocabulary, definedSyms, settings)
    }
  
  //////////////////////////////////////////////////////////////////////////////
  
  val subtrees : Seq[ProofTree] = List()
  
  override def toString : String = {
    "^ " + closingConstraint + "\n" +
    ". Facts: " + facts + "\n" +
    ". Clauses: " + clauses + "\n" +
    ". Tasks: " + tasks + "\n" +
    ". Eliminated constants: " + eliminatedConstants + "\n" +
    ". Free constants " + (if (fixedConstantFreedom) "(fixed)" else "(not fixed)") +
                     ": " + constantFreedom + "\n" +
    ". Defined constants: " + definedSyms
  }
}
