package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Term, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.Conjunction
import ap.terfor.inequalities.InEqConj
import ap.terfor.equations.NegEquationConj
import ap.terfor.arithconj.{ArithConj, ModelFinder}
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Logic, Debug, Seqs, PlainRange}

import scala.collection.mutable.ArrayBuffer

/**
 * Task for eliminating inequalities using the equivalence from the Omega-test
 */
case object OmegaTask extends EagerTask {

  private val AC = Debug.AC_OMEGA

  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val ac = goal.facts.arithConj
    val order = goal.order
    
    selectInequalities(goal) match {
      case Some((lowerBounds, upperBounds, remainder)) => {
        val eliminatedConstant = lowerBounds(0).leadingTerm.asInstanceOf[ConstantTerm]
        
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(AC,
                        Logic.exists(for (lc <- lowerBounds.elements)
                                     yield (!lc.leadingCoeff.abs.isOne)) &&
                        Logic.exists(for (lc <- upperBounds.elements)
                                     yield (!lc.leadingCoeff.abs.isOne)) &&
                        !(goal.tasks.taskInfos.constants contains eliminatedConstant) &&
                        !(ac.positiveEqs.constants contains eliminatedConstant) &&
                        !(ac.negativeEqs.constants contains eliminatedConstant) &&
                        !(goal.facts.predConj.constants contains eliminatedConstant) &&
                        !(goal.clauses.constants contains eliminatedConstant))
        ////////////////////////////////////////////////////////////////////////
          
        val newGoals = new ArrayBuffer[ProofTree]
          
        // the splinters
        for (eq <- splinterEqs(lowerBounds, upperBounds, order))
          newGoals += ptf.updateGoal(goal.formulaTasks(eq), goal)

        // the dark shadow
        val darkShadowConj = darkShadow(lowerBounds, upperBounds, order)
        val darkShadowGoal = ptf.updateGoal(goal.facts.updateInEqs(remainder, order),
                                            goal.formulaTasks(darkShadowConj.negate),
                                            goal)
        
        // we add a <code>ModelFinder</code> so that a witness for the eliminated
        // constant can be constructed
        val eliminatedInEqs =
          ArithConj.conj(InEqConj(lowerBounds.elements ++ upperBounds.elements,
                                  order), order)
        newGoals += ptf.eliminatedConstant(darkShadowGoal,
                                           eliminatedConstant,
                                           new ModelFinder (eliminatedInEqs,
                                                            eliminatedConstant),
                                           goal.vocabulary)
        
        ptf.and(newGoals, goal.vocabulary)
      }
      
      case None => // there are no inequalities that could be eliminated
                   ptf updateGoal goal
    }
    
  }

  /**
   * Determine whether this task might causes proof splitting
   */
  def splittingNecessary(goal : Goal) : Boolean =
    selectInequalities(goal) != None

  /**
   * Generate the dark shadow (i.e., a conjunction of inequalities).
   */
  private def darkShadow(lowerBounds : Seq[LinearCombination],
                         upperBounds : Seq[LinearCombination],
                         order : TermOrder) : Conjunction = {
    val infs =
      for (geq <- lowerBounds.elements; leq <- upperBounds.elements) yield {
        val leqLC = leq.leadingCoeff // negative
        val geqLC = geq.leadingCoeff // positive
        val correction = (leqLC + IdealInt.ONE) * (geqLC - IdealInt.ONE)
   
        LinearCombination.sum(Array((-leqLC, geq), (geqLC, leq),
                                    (correction, LinearCombination.ONE)),
                              order)
    }
    Conjunction.conj(InEqConj(infs, order), order)
  }

  /**
   * Generate the equations that describe the splinters.
   */
  private def splinterEqs(lowerBounds : Seq[LinearCombination],
                          upperBounds : Seq[LinearCombination],
                          order : TermOrder) : Iterator[Conjunction] = {
    
    // the leading coefficients of the upper bounds are all negative
    val m = -IdealInt.min(for (lc <- upperBounds.elements) yield lc.leadingCoeff)
    
    for (lc <- lowerBounds.elements;
         k <- PlainRange((((m - IdealInt.ONE) * lc.leadingCoeff - m) / m).intValueSafe + 1)
              .elements)
    yield Conjunction.conj(NegEquationConj(lc + IdealInt(-k), order), order)
  }

  /**
   * Select the inequalities that are to be eliminated. The result consists of
   * two sets of inqualities and the conjunction of the remaining, non-eliminated
   * equalities.
   */
  private def selectInequalities(goal : Goal)
        : Option[(Seq[LinearCombination], Seq[LinearCombination], InEqConj)] = {
    
    val facts = goal.facts
    val order = goal.order
    
    val constantsSeen = new scala.collection.mutable.HashSet[ConstantTerm]
    constantsSeen ++= facts.arithConj.positiveEqs.constants
    constantsSeen ++= facts.arithConj.negativeEqs.constants
    constantsSeen ++= facts.predConj.constants
    constantsSeen ++= goal.clauses.constants
    
    for (lc <- facts.arithConj.inEqs) {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(OmegaTask.AC, lc.leadingTerm.isInstanceOf[ConstantTerm])
      //////////////////////////////////////////////////////////////////////////
      
      val lt = lc.leadingTerm.asInstanceOf[ConstantTerm]
      
      // we cannot eliminate inequalities whose leading term also occurs as
      // non-leading term in other inequalities or in non-inequalities
      if ((goal eliminates lt) && !(constantsSeen contains lt)) {
        // select inequalities with the leading coefficient lt
        val lowerBounds = new ArrayBuffer[LinearCombination]        
        val upperBounds = new ArrayBuffer[LinearCombination]        
        val remainingInEqs = new ArrayBuffer[LinearCombination]
        
        for (lc <- facts.arithConj.inEqs) {
          if (lt == lc.leadingTerm) {
            if (lc.leadingCoeff.signum > 0)
              lowerBounds += lc
            else
              upperBounds += lc
          } else {
            remainingInEqs += lc
          }
        }
        
        return Some(lowerBounds, upperBounds, InEqConj(remainingInEqs, goal.order))
      }
      
      constantsSeen ++= lc.constants
    }
    
    None
  }
  
}
