package ap.proof.unificationOpt;

import ap.terfor.{ConstantTerm, TermOrder}
import ap.terfor.conjunctions.Quantifier
import ap.terfor.equations.EquationConj
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.substitutions._
import ap.proof.goal.Goal
import ap.util.{Debug, Logic, Seqs}

import scala.util.Sorting

object ConstantOrder {
  
  private val AC = Debug.AC_UNIFICATION_OPT
  
  abstract class ConstantKind
  case object GlobalConstant extends ConstantKind
  case object UniversalConstant extends ConstantKind
  case object ExistentialConstant extends ConstantKind
  case object SubstConstant extends ConstantKind
  
  def fromProof(p : ProofTree) : ConstantOrder = {
    val kinds = new scala.collection.mutable.HashMap[BoundConstant, ConstantKind]
    buildMaps(p, List(0), kinds)
    
    new ConstantOrder {
      def kindOf(c : BoundConstant) : ConstantKind =
        if (c.position.isEmpty) GlobalConstant else kinds(c)
      lazy val knownConstants : scala.collection.Set[BoundConstant] = kinds.keySet
    }
  }
  
  private def buildMaps
       (p : ProofTree,
        currentPos : List[Int],
        kinds : scala.collection.mutable.Map[BoundConstant, ConstantKind]) : Unit = {
    def addConst(c : BoundConstant, kind : ConstantKind) = {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(AC, !(kinds contains c))
      //////////////////////////////////////////////////////////////////////////
      kinds += (c -> kind)
    }
    
    p match {
      case goal : Goal => {
        val consts = substitutedConstants(goal.definedSyms)
        ////////////////////////////////////////////////////////////////////////
        // make sure that the substituted constants are actually removed from
        // constraints, i.e., the substitution is idempotent
        Debug.assertInt(AC, Logic.forall(for (c <- consts.elements) yield {
          val order = TermOrder.EMPTY.extend(c, Set())
          val eq = EquationConj(LinearCombination(c, order), order)
          Seqs.disjoint(goal.definedSyms(eq).constants, consts)
        }))
        ////////////////////////////////////////////////////////////////////////
        for (c <- consts) addConst(BoundConstant(c, currentPos), SubstConstant)
      }
      
      case AndTree(left, right) => {
        buildMaps(left, 0 :: currentPos, kinds)
        buildMaps(right, 1 :: currentPos, kinds)
      }
      
      case QuantifiedTree(q, consts, subtree) => {
        val kind = if (q == Quantifier.ALL) UniversalConstant else ExistentialConstant
        for (c <- consts) addConst(BoundConstant(c, currentPos), kind)
        buildMaps(subtree, 0 :: currentPos, kinds)
      }
      
      case ProofTreeOneChild(subtree) =>
        buildMaps(subtree, currentPos, kinds)
    }
  }
  
  private def substitutedConstants(s : Substitution) : Set[ConstantTerm] = s match {
    case IdentitySubst() => Set()
    case ConstantSubst(c, _) => Set(c)
    case PseudoConstantSubst(_, c, _) => Set(c)
    case ComposeSubsts(s1, s2) =>
      substitutedConstants(s1) ++ substitutedConstants(s2)
  }
}

/**
 * Class for collecting the constants that have been introduced in a proof, as
 * well as the partial ordering given by the order of rule applications. A
 * constant c is strictly smaller than another constant d if both constants have
 * been introduced on the same branch, and if c has been introduce earlier 
 * (farther outside) than d
 */
trait ConstantOrder {

  def kindOf(c : BoundConstant) : ConstantOrder.ConstantKind
  
  val knownConstants : scala.collection.Set[BoundConstant]

  /**
   * Find the innermost bound occurrence of the constant <code>c</code>. If
   * no bound occurrence is known, <code>c</code> is assumed to be global
   */
  def asBound(c : ConstantTerm, pos : List[Int]) : BoundConstant = {
    val cur = BoundConstant(c, pos)
    if (pos.isEmpty || (knownConstants contains cur))
      cur
    else
      asBound(c, pos.tail)
  }

  def asBound(cs : Iterable[ConstantTerm], pos : List[Int]) : Iterable[BoundConstant] =
    for (c <- cs) yield asBound(c, pos)

  def asBound(cs : Set[ConstantTerm], pos : List[Int]) : Set[BoundConstant] =
    for (c <- cs) yield asBound(c, pos)

  def maximumConstants(consts : Iterable[BoundConstant]) : Seq[BoundConstant] =
    if (consts.isEmpty) {
      List()
    } else {
      val constsArray = Seqs toArray consts
      Sorting.stableSort(constsArray, (a:BoundConstant, b:BoundConstant) => a >= b)
      val res = constsArray takeWhile (constsArray(0) <= _)
      res
    }
  
}
