package ap.proof.unificationOpt;

import ap.proof.goal.{Goal, TaskManager}
import ap.terfor.ConstantTerm
import ap.terfor.conjunctions.{Conjunction, Quantifier, NegatedConjunctions}
import ap.terfor.substitutions._
import ap.util.{FilterIt, Debug, Logic, Seqs}

object QuasiUniConstants {

  private val AC = Debug.AC_UNIFICATION_OPT

  private type Dependencies =
    scala.collection.Map[BoundConstant, Set[BoundConstant]]  
  private type MutableDependencies =
    scala.collection.mutable.Map[BoundConstant, Set[BoundConstant]]  

  private def transClosure(deps : Dependencies) : Dependencies = {
    val newDeps = new scala.collection.mutable.HashMap[BoundConstant, Set[BoundConstant]]

    var changed = false
    for ((c, ds) <- deps) {
      val newDs = ds ++ (for (d <- ds.elements; e <- deps(d).elements) yield e)
      if (newDs.size != ds.size) changed = true
      newDeps += (c -> newDs)
    }
    
    if (changed)
      transClosure(newDeps)
    else
      newDeps
  }

  /**
   * Run some destructive operation until a given measure does no longer change
   */
  private def fix[A](measure : => A, op : => Unit) : Unit =
    while (true) {
      val oldMeasure = measure
      op
      if (oldMeasure == measure) return
    }
  
}

class QuasiUniConstants(tree : ProofTree) {

  import QuasiUniConstants.{Dependencies, MutableDependencies}
  
  private val order = ConstantOrder fromProof tree

  private def quCandidate(c : BoundConstant) : Boolean =
    (order kindOf c) == ConstantOrder.UniversalConstant

  private val deps = dependencies(tree)
  
  private def closeUnderDeps(consts : Set[BoundConstant]) : Set[BoundConstant] =
    consts ++ (for (c <- consts; d <- deps.getOrElse(c, Set())) yield d)
  
  val result : scala.collection.Set[BoundConstant] = {
    val quConsts = new scala.collection.mutable.HashSet[BoundConstant]
    quConsts ++= FilterIt(order.knownConstants.elements, quCandidate _)
    QuasiUniConstants.fix(quConsts.size, reduceQUCSet(tree, List(0), quConsts))
    quConsts
  }

  val simplerProof = simplifyProof(tree, List(0), result)
  
  //////////////////////////////////////////////////////////////////////////////
  
  private def simplifyProof(p : ProofTree,
                            currentPos : List[Int],
                            quConsts : scala.collection.Set[BoundConstant]) : ProofTree =
    p match {
      case goal : Goal => {
        val remainingFors = unshieldedPart(goal.closingConstraint, currentPos, quConsts).negate
        val newFacts = remainingFors.updateNegatedConjs(NegatedConjunctions.TRUE, goal.order)
        Goal (newFacts,
              remainingFors.negatedConjs,
              TaskManager.EMPTY,
              goal.age,
              goal.eliminatedConstants,
              goal.vocabulary, goal.definedSyms,
              goal.settings)
      }
                              
      case AndTree(left, right) =>
        p.asInstanceOf[AndTree].update(simplifyProof(left, 0 :: currentPos, quConsts),
                                       simplifyProof(right, 1 :: currentPos, quConsts),
                                       p.constantFreedom)
                              
      case p : QuantifiedTree =>
        p.update(simplifyProof(p.subtree, 0 :: currentPos, quConsts), p.constantFreedom)

      case WeakenTree(disjunct, subtree) =>
        WeakenTree(simplifyProof(subtree, currentPos, quConsts),
                   unshieldedPart(disjunct, currentPos, quConsts),
                   p.vocabulary)
    }    
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Starting with an initial set of constants, iteratively remove all constants
   * that violate the conditions on a set of quasi-universal constants
   */
  private def reduceQUCSet(p : ProofTree,
                           currentPos : List[Int],
                           quConsts : scala.collection.mutable.Set[BoundConstant]) : Unit = {
    def removeNonUniversalConstants(c : Conjunction) : Unit =
      QuasiUniConstants.fix(quConsts.size,
                            quConsts --= nonUniversalConstants(c, currentPos, quConsts))
    
    p match {
    case goal : Goal => {
      val oldS = quConsts.size
      removeNonUniversalConstants(goal.closingConstraint)
      if (quConsts.size < oldS) {
        println("Number of quasi-universal constants dropped from " + oldS + " to " + quConsts.size)
        println(unshieldedPart(goal.closingConstraint, currentPos, quConsts))
      }
    }
    
    case AndTree(left, right) => {
      reduceQUCSet(left, 0 :: currentPos, quConsts)
      reduceQUCSet(right, 1 :: currentPos, quConsts)
    }
    
    case QuantifiedTree(q, consts, subtree) => {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(QuasiUniConstants.AC,
                      q != Quantifier.EX ||
                      Seqs.disjoint(order.asBound(consts, currentPos), quConsts))
      //////////////////////////////////////////////////////////////////////////
      reduceQUCSet(subtree, 0 :: currentPos, quConsts)
    }
    
    case WeakenTree(disjunct, subtree) => {
      removeNonUniversalConstants(disjunct)
      reduceQUCSet(subtree, currentPos, quConsts)
    }
    
    }
  }
  
  private def nonUniversalConstants(c : Conjunction,
                                    currentPos : List[Int],
                                    quConsts : scala.collection.Set[BoundConstant])
                                                        : Set[BoundConstant] =
    closeUnderDeps(order.asBound(unshieldedPart(c, currentPos, quConsts).constants,
                                 currentPos))
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Determine the dependencies between universal constants that exist due
   * to substitutions in the proof tree. E.g., if there is a substitution node
   * [c/t] in the tree, then the constant c depends on all constants in t.
   */
  private def dependencies(p : ProofTree) : Dependencies = {
    val deps = new scala.collection.mutable.HashMap[BoundConstant, Set[BoundConstant]]    
    depsForProof(p, List(0), deps)
    QuasiUniConstants transClosure deps
  }
  
  private def depsForProof(p : ProofTree,
                           currentPos : List[Int],
                           deps : MutableDependencies) : Unit =
    p match {
      case QuantifiedTree(_, _, subtree) =>
        depsForProof(subtree, 0 :: currentPos, deps)
      case ProofTreeOneChild(subtree) =>
        depsForProof(subtree, currentPos, deps)
      case AndTree(left, right) => {
        depsForProof(left, 0 :: currentPos, deps)
        depsForProof(right, 1 :: currentPos, deps)
      }
      case goal : Goal =>
        depsForSubst(goal.definedSyms, currentPos, deps)
    }

  private def depsForSubst(s : Substitution,
                           currentPos : List[Int],
                           deps : MutableDependencies) : Unit = {
    def addDeps(c : ConstantTerm, ds : Iterable[ConstantTerm]) = {
      val bc = order.asBound(c, currentPos)
      val oldDs = deps.getOrElse(bc, Set())
      deps += (bc -> (oldDs ++ FilterIt(for (d <- ds.elements)
                                         yield order.asBound(d, currentPos),
                                         quCandidate _)))
    }
    
    s match {
    case IdentitySubst() =>
      // nothing
    case ConstantSubst(c, t) =>
      addDeps(c, t.constants)
    case PseudoConstantSubst(_, c, t) =>
      addDeps(c, t.constants)
    case ComposeSubsts(s1, s2) => {
      depsForSubst(s1, currentPos, deps)
      depsForSubst(s2, currentPos, deps)
    }
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Determine whether <code>c</code> is a formula of the shape
   * <code> a*x + t = 0 & phi </code>, where <code>a != 0</code> and
   * <code>x</code> is a quasi-universal constant that is maximal in
   * <code> a*x + t = 0 </code>.
   */
  private def isShielded(c : Conjunction,
                         currentPos : List[Int],
                         quConsts : scala.collection.Set[BoundConstant]) : Boolean = {
    ////////////////////////////////////////////////////////////////////////////
    // assume that c is properly simplified
    Debug.assertPre(QuasiUniConstants.AC, c.arithConj.inEqs.equalityInfs.isTrue)
    ////////////////////////////////////////////////////////////////////////////

    // TODO: generalise this
    Logic.exists(for (lc <- c.arithConj.positiveEqs.elements)
                 yield {
                         val boundCs = order.asBound(lc.constants, currentPos)
                         !Seqs.disjointSeq(quConsts, order maximumConstants boundCs)
                       })
  }

  private def isShieldedNeg(c : Conjunction,
                            currentPos : List[Int],
                            quConsts : scala.collection.Set[BoundConstant]) : Boolean =
    isShielded(c.negate, currentPos, quConsts)

  /**
   * Determine the (disjunctively connected) unshielded part of a formula
   */
  private def unshieldedPart(c : Conjunction,
                             currentPos : List[Int],
                             quConsts : scala.collection.Set[BoundConstant])
                                                         : Conjunction =
    if (isShielded(c, currentPos, quConsts))
      Conjunction.FALSE
    else if (c.isNegatedConjunction)
      Conjunction.conj(FilterIt(c.negatedConjs(0).elements,
                                (d:Conjunction) => !isShieldedNeg(d, currentPos, quConsts)),
                       c.order).negate
    else
      // TODO: generalise?
      c

}
