package ap.terfor.preds;

import ap.terfor.substitutions.VariableShiftSubst
import ap.util.{Debug, Seqs, LazyMappedSet, UnionSet}

object ReduceWithPredLits {
  
  private val AC = Debug.AC_PROPAGATION
  
  def apply(conj : PredConj, order : TermOrder) : ReduceWithPredLits = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    new ReduceWithPredLits(conj.positiveLitsAsSet, conj.negativeLitsAsSet,
                           conj.predicates, order)
  }
    
}

class ReduceWithPredLits private (positiveLits : scala.collection.Set[Atom],
                                  negativeLits : scala.collection.Set[Atom],
                                  preds : scala.collection.Set[Predicate],
                                  order : TermOrder) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ReduceWithPredLits.AC,
                   preds == Set() ++ (for (a <- positiveLits) yield a.pred) ++
                                     (for (a <- negativeLits) yield a.pred))
  //////////////////////////////////////////////////////////////////////////////

  def addLits(furtherLits : PredConj) : ReduceWithPredLits =
    if (furtherLits.isTrue)
      this
    else
      new ReduceWithPredLits(UnionSet(positiveLits, furtherLits.positiveLitsAsSet),
                             UnionSet(negativeLits, furtherLits.negativeLitsAsSet),
                             UnionSet(preds, furtherLits.predicates),
                             order)

  /**
   * Create a <code>ReduceWithEqs</code> that can be used underneath
   * <code>num</code> binders. The conversion of de Brujin-variables is done on
   * the fly, which should give a good performance when the resulting
   * <code>ReduceWithEqs</code> is not applied too often (TODO: caching)
   */
  def passQuantifiers(num : Int) : ReduceWithPredLits = {
    val upShifter = VariableShiftSubst.upShifter[Atom](num, order)
    val downShifter = VariableShiftSubst.downShifter[Atom](num, order)
    new ReduceWithPredLits(new LazyMappedSet(positiveLits, upShifter, downShifter),
                           new LazyMappedSet(negativeLits, upShifter, downShifter),
                           preds, order)
  }

  /**
   * Determine whether a formula that contains the given predicates might be
   * reducible by this reducer
   */
  def reductionPossible(conj : PredConj) : Boolean =
    !Seqs.disjoint(preds, conj.predicates)

  def apply(conj : PredConj) : PredConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithPredLits.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    if (reductionPossible(conj)) {
      // TODO: should be done using binary search
      if (!Seqs.disjointSeq(positiveLits, conj.negativeLits) ||
          !Seqs.disjointSeq(negativeLits, conj.positiveLits))
        // a contradiction has been found
        PredConj.FALSE(conj)
      else
        conj.updateLitsSubset(conj.positiveLits filter ((a) => !(positiveLits contains a)),
                              conj.negativeLits filter ((a) => !(negativeLits contains a)),
                              order)
    } else {
      conj
    }
  }
  
}
